/* Convert AppleDouble into MacBinary */
/*         Copyright(c)2003 by T.Toda */
#include <stdio.h>
#include <string.h>

#ifdef	MSDOS
#define	READMODE	"rb"
#define	WRITEMODE	"wb"
#else
#ifdef	WIN32
#define	READMODE	"rb"
#define	WRITEMODE	"wb"
#else
#define	READMODE	"r"
#define	WRITEMODE	"w"
#endif
#endif

/* Offset of HFS Date from AppleDouble Date       */
/* Value of Seconds of 96 years from 1904 to 2000 */
#define	DATEOFFSET	((96UL / 4) * (365UL * 4 + 1) * 24UL * 60UL * 60UL)

static	void	fputhex(unsigned long, int, FILE *);
static	unsigned long	makelong(unsigned char *);
static	unsigned short	makeshort(unsigned char *);
static	void	divlong(unsigned char *, unsigned long);
static	void	divshort(unsigned char *, unsigned short);
static	unsigned short	makecrc(unsigned char *, unsigned short);

static	char	*idtext[]
	= {"Unknown", "Data Fork", "Resource Fork", "Real Name", 
	   "Comment", "BW Icon", "Color Icon", "File Info", 
	   "File Dates", "Finder Info", "Mac File Info", "ProDos File Info", 
	   "MSDOS File Info", "Short Name", "AFP File Info", "Directory ID"};

int	main(int argc, char **argv, char **envp)
{
	FILE	*fphead, *fpdata, *fpout;
	struct{
		unsigned char	oldversion;
		unsigned char	fnamelen;
		char	filename[63];
		char	filetype[4];
		char	creator[4];
		unsigned char	finderflag1;
		unsigned char	zerofill1;
		unsigned char	vertical[2];
		unsigned char	horizontal[2];
		unsigned char	folderid[2];
		unsigned char	protected;
		unsigned char	zerofill2;
		unsigned char	datafolk[4];
		unsigned char	resourcefolk[4];
		unsigned char	created[4];
		unsigned char	modified[4];
		unsigned char	infocomment[2];	/* for MacBinary II */
		unsigned char	finderflag2;	/* for MacBinary II */
		unsigned char	signature[4];	/* for MacBinary III */
		unsigned char	fnamescript;	/* for MacBinary III */
		unsigned char	extendedflag;	/* for MacBinary III */
		unsigned char	reserved[8];
		unsigned char	unpackedsize[4];	/* for MacBinary II */
		unsigned char	secondheader[2];	/* for MacBinary II */
		unsigned char	version;	/* for MacBinary II */
		unsigned char	minversion;	/* for MacBinary II */
		unsigned char	crc[2];	/* for MacBinary II */
		unsigned char	osid[2];	/* for MacBinary I */
	}	macbinheader;
	struct{
		unsigned char	magic[4];
		unsigned char	version[4];
		char	filler[16];
		unsigned char	entries[2];
	}	appleheader;
	struct{
		unsigned char	id[4];
		unsigned char	offset[4];
		unsigned char	length[4];
	}	appleentry;
	unsigned long	i_magic, i_id;
	unsigned short	i_entries, i;
	unsigned long	i_offset[16], i_length[16];
	unsigned long	datalength, l;
	int	c;

	if( argc > 4 || argc < 2 ){
		fputs("Usage : ", stderr);
		fputs(argv[0], stderr);
		fputs(" datafile [headerfile [outfile]]\n", stderr);
		return	4;
	}
	if(( fpdata = fopen(argv[1],READMODE)) == NULL ){
		fputs("Can't Open '", stderr);
		fputs(argv[1], stderr);
		fputs("'.\n", stderr);
		return	8;
	}
	if( argc < 3 ){
		fphead = stdin;
	}
	else if(( fphead = fopen(argv[2],READMODE)) == NULL ){
		fputs("Can't Open '", stderr);
		fputs(argv[2], stderr);
		fputs("'.\n", stderr);
		return	8;
	}
	if( argc < 4 ){
		fpout = stdout;
	}
	else if(( fpout = fopen(argv[3],WRITEMODE)) == NULL ){
		fputs("Can't Open '",stderr);
		fputs(argv[3],stderr);
		fputs("'.\n",stderr);
		return	8;
	}

	memset(&macbinheader, 0, sizeof(macbinheader));

	fseek(fpdata, 0L, 2);
	datalength = ftell(fpdata);
	rewind(fpdata);
	divlong(macbinheader.datafolk, datalength);

	fread(&appleheader, sizeof(appleheader), 1, fphead);
	i_entries = makeshort(appleheader.entries);
	i_magic = makelong(appleheader.magic);
	if( i_magic != 0x00051607 ){
		fputs("Invalid AppleDouble Identifier : ", stderr);
		fputhex(i_magic, 8, stderr);
		putc('\n', stderr);
		return	8;
	}

	for( i = 0 ; i < 16 ; i++ ){
		i_offset[i] = 0;
	}
	for( i = 0 ; i < i_entries ; i++ ){
		fread(&appleentry, sizeof(appleentry), 1, fphead);
		i_id = makelong(appleentry.id);
		if( i_id > 16 ){
			fputs("Warning: Unknown Entry ID", stderr);
			fputhex(i_id, 8, stderr);
			putc('\n', stderr);
		}
		else if( i_offset[i_id] ){
			fputs("Warning: Duplicated Entry ID", stderr);
			fputhex(i_id, 1, stderr);
			putc('(', stderr);
			fputs(idtext[i_id], stderr);
			fputs(")\n", stderr);
		}
		else{
			i_offset[i_id] = makelong(appleentry.offset);
			i_length[i_id] = makelong(appleentry.length);
		}
	}

	/* 3: Real Name */
	if( i_offset[3] ){
		fseek(fphead, (signed long)i_offset[3], 0);
		if( i_length[3] > 63 )i_length[3] = 63;
		macbinheader.fnamelen = (unsigned char)i_length[3];
		fread(macbinheader.filename, (unsigned short)i_length[3], 1, fphead);
	}

	/* 9: Finder Info */
	/* Consists of 16-Byte 'Finder Info' and      */
	/* 16-Byte 'Exteneded Finder Info' (ignored). */
	/* 'Finder Info' consists of                  */
	/* 'filetype', 'creator', 'finderflag1',      */
	/* 'finderflag1', 'zerofill1' ('finderflag2') */
	/* 'vertical', 'horizontal', and 'folderid'.  */
	/* 'finderflag2' is added in MacBinary II and */
	/* its counterpart 'zerofill1' should be      */
	/* filled with 0 for compatibility.           */
	if( i_offset[9] ){
		fseek(fphead, (signed long)i_offset[9], 0);
		fread(macbinheader.filetype, 16, 1, fphead);
		macbinheader.finderflag2 = macbinheader.zerofill1;
		macbinheader.zerofill1 = 0;
	}

	/* 8: File Dates */
	if( i_offset[8] ){
		fseek(fphead, (signed long)i_offset[8], 0);
		fread(macbinheader.created, 8, 1, fphead);
		divlong(macbinheader.created, makelong(macbinheader.created) + DATEOFFSET);
		divlong(macbinheader.modified, makelong(macbinheader.modified) + DATEOFFSET);
	}

	/* 10: Mac File Info */
	if( i_offset[10] ){
		fseek(fphead, (signed long)i_offset[10], 0);
		fread(&macbinheader.protected, 1, 1, fphead);
	}

	/* 2: Resource Fork */
	if( i_offset[2] ){
		divlong(macbinheader.resourcefolk, i_length[2]);
	}

	macbinheader.version = macbinheader.minversion = 0x81;
	divshort(macbinheader.crc, makecrc((unsigned char *)&macbinheader, 0x7e));

	fwrite(&macbinheader, sizeof(macbinheader), 1, fpout);

	for( l = 0 ; l < datalength ; l++ ){
		if( (c = getc(fpdata)) == EOF ){
			fputs("Error: Data File Shorter than Pre-Seeked\n", stderr);
			return	12;
		}
		putc(c, fpout);
	}
	if( getc(fpdata) != EOF ){
		fputs("Warning: Data File Longer than Pre-Seeked\n", stderr);
	}
	while( l % 128 ){
		putc(0, fpout);
		l++;
	}

	if( i_offset[2] ){
		fseek(fphead, (signed long)i_offset[2], 0);
		for( l = 0 ; l < i_length[2] ; l++ ){
			if( (c = getc(fphead)) == EOF ){
				fputs("Warning: End of File in Estimated Resource Fork\n", stderr);
				break;
			}
			putc(c, fpout);
		}
		while( l % 128 ){
			putc(0, fpout);
			l++;
		}
	}
	return	0;
}

#define	CRC16_POLYNOMIAL	0x1021;
static	unsigned short	makecrc(unsigned char *dataseq, unsigned short length)
{
	unsigned short	result = 0;
	unsigned short	i, byte, bit;

	for( i = 0 ; i < length ; i++ ){
		byte = dataseq[i];
		for( bit = 0 ; bit < 8 ; bit++ ){
			if( result & 0x8000 ){
				result <<= 1;
				result ^= CRC16_POLYNOMIAL;
			}
			else{
				result <<= 1;
			}
			if( byte & 0x80 ){
				result ^= 1;
			}
			byte <<= 1;
		}
	}
	return	result;
}

static void	fputhex(unsigned long c, int l, FILE *fp)
{
	static	char	const	hexnum[] = "0123456789ABCDEF";
	char	result[9];

	if(l > 8)l = 8;
	if(l < 1)l = 1;
	result[l] = 0;
	result[--l] = hexnum[c % 0x10];
	while( l-- )
		result[l] = hexnum[(c >>= 4) % 0x10];
	fputs(result, fp);
}

static	unsigned long	makelong(unsigned char *buffer)
{
	return	( ((long)buffer[0] << 24 ) | 
	      	  ((long)buffer[1] << 16 ) | 
	      	  ((long)buffer[2] <<  8 ) | 
	      	  ((long)buffer[3] ) );
}

static	unsigned short	makeshort(unsigned char *buffer)
{
	return	( ((long)buffer[0] <<  8 ) | 
	      	  ((long)buffer[1] ) );
}

static	void	divlong(unsigned char *buffer, unsigned long data)
{
	buffer[0] = (unsigned char)(( data >> 24 ) & 0x000000ff );
	buffer[1] = (unsigned char)(( data >> 16 ) & 0x000000ff );
	buffer[2] = (unsigned char)(( data >>  8 ) & 0x000000ff );
	buffer[3] = (unsigned char)(  data         & 0x000000ff );
}

static	void	divshort(unsigned char *buffer, unsigned short data)
{
	buffer[0] = (unsigned char)(( data >>  8 ) & 0x000000ff );
	buffer[1] = (unsigned char)(  data         & 0x000000ff );
}
