/* Divide AppleSingle File         */
/* into a Set of AppleDouble Files */
/*      Copyright(c)2003 by T.Toda */
#include <stdio.h>

#ifdef	MSDOS
#define	READMODE	"rb"
#define	WRITEMODE	"wb"
#else
#ifdef	WIN32
#define	READMODE	"rb"
#define	WRITEMODE	"wb"
#else
#define	READMODE	"r"
#define	WRITEMODE	"w"
#endif
#endif

static	void	fputhex(unsigned long, int, FILE *);
static	unsigned long	makelong(unsigned char *);
static	unsigned short	makeshort(unsigned char *);
static	void	divlong(unsigned char *, unsigned long);
static	void	divshort(unsigned char *, unsigned short);

int	main(int argc, char **argv, char **envp)
{
	FILE	*fphead, *fpdata, *fpin;
	struct{
		unsigned char	magic[4];
		unsigned char	version[4];
		char	filler[16];
		unsigned char	entries[2];
	}	header;
	struct{
		unsigned char	id[4];
		unsigned char	offset[4];
		unsigned char	length[4];
	}	entry;
	unsigned long	i_magic, i_offset, i_length;
	unsigned short	i_entries;
	unsigned short	i, i_datafork;
	unsigned long	maxend, bytewrite, d_offset, d_length;
	int	c;

	if( argc > 4 || argc < 3 ){
		fputs("Usage : ", stderr);
		fputs(argv[0], stderr);
		fputs(" infile headerfile [datafile]\n", stderr);
		return	4;
	}
	if( (fpin = fopen(argv[1], READMODE)) == NULL ){
		fputs("Can't Open '", stderr);
		fputs(argv[1], stderr);
		fputs("'.\n", stderr);
		return	8;
	}
	if( (fphead = fopen(argv[2], READMODE)) != NULL ){
		fputs("Destinate File '", stderr);
		fputs(argv[2], stderr);
		fputs("' Exists. Overwrite? (Y/N)", stderr);
		c = getchar();
		if( (c != 'Y') && (c != 'y') ){
			return	8;
		}
		while( getchar() != '\n' );
	}
	fclose(fphead);
	if( (fphead = fopen(argv[2], WRITEMODE)) == NULL ){
		fputs("Can't Open '", stderr);
		fputs(argv[2], stderr);
		fputs("'.\n", stderr);
		return	8;
	}
	if( argc < 4 ){
		fpdata = stdout;
	}
	else{
		if( (fpdata = fopen(argv[3], READMODE)) != NULL ){
			fputs("Destinate File '", stderr);
			fputs(argv[3], stderr);
			fputs("' Exists. Overwrite? (Y/N)", stderr);
			c = getchar();
			if( (c != 'Y') && (c != 'y') ){
				return	8;
			}
			while( getchar() != '\n' );
		}
		fclose(fpdata);
		if( (fpdata = fopen(argv[3], WRITEMODE)) == NULL ){
			fputs("Can't Open '",stderr);
			fputs(argv[3],stderr);
			fputs("'.\n",stderr);
			return	8;
		}
	}

	fread(&header, sizeof(header), 1, fpin);
	maxend = sizeof(header);
	i_entries = makeshort(header.entries);
	i_magic = makelong(header.magic);
	if( i_magic != 0x00051600 ){
		fputs("Invalid AppleSingle Identifier : ", stderr);
		fputhex(i_magic, 8, stderr);
		putc('\n', stderr);
		return	8;
	}

	i_datafork = i_entries;
	for( i = 0 ; i < i_entries ; i++ ){
		fread(&entry, sizeof(entry), 1, fpin);
		if( makelong(entry.id) == 1 ){
			if( i_datafork != i_entries ){
				fputs("Warning: More than one Data Forks\n", stderr);
			}
			else{
				i_datafork = i;
				d_length = makelong(entry.length);
				d_offset = makelong(entry.offset);
			}
		}
	}
	if( i_datafork == i_entries ){
		fputs("Warning: No Data Fork was found\n", stderr);
	}

	rewind(fpin);
	fread(&header, sizeof(header), 1, fpin);
	if( (i_entries != makeshort(header.entries)) ||
	    (i_magic != makelong(header.magic)) ){
		fputs("Error: Re-Seeked Header does not match with Pre-Seeked\n", stderr);
		return	8;
	}
	header.magic[3] = 0x07;
	divshort(header.entries, i_entries - 1);
	fwrite(&header, sizeof(header), 1, fphead);
	bytewrite = sizeof(header);

	for( i = 0 ; i < i_entries ; i++ ){
		fread(&entry, sizeof(entry), 1, fpin);
		if( i == i_datafork )continue;
		i_length = makelong(entry.length);
		i_offset = makelong(entry.offset);
		if( i_offset >= d_offset ){
			if( i_offset < d_offset + d_length ){
				fputs("Warning: Estimated Offset is in Data Fork Area\n", stderr);
			}
			else{
				i_offset -= d_length;
			}
		}
		i_offset -= sizeof(entry);
		divlong(entry.offset, i_offset);
		if( maxend < i_offset + i_length ){
			maxend = i_offset + i_length;
		}
		fwrite(&entry, sizeof(entry), 1, fphead);
		bytewrite += sizeof(entry);
	}

	while( bytewrite < d_offset - sizeof(entry) ){
		if( (c = getc(fpin)) == EOF ){
			fputs("Error: End of File before Estimated Data Fork\n", stderr);
			return	12;
		}
		bytewrite++;
		putc(c, fphead);
	}

	while( d_length-- ){
		if( (c = getc(fpin)) == EOF ){
			fputs("Error: Data Fork Shorter than Estimated\n", stderr);
			return	12;
		}
		putc(c, fpdata);
	}

	while( bytewrite < maxend ){
		if( (c = getc(fpin)) == EOF ){
			fputs("Error: Header Shorter than Estimated\n", stderr);
			return	12;
		}
		bytewrite++;
		putc(c, fphead);
	}
	if( getc(fpin) != EOF ){
		fputs("Warning: Header Longer than Estimated\n", stderr);
	}
	return	0;
}

static void	fputhex(unsigned long c, int l, FILE *fp)
{
	static	char	const	hexnum[] = "0123456789ABCDEF";
	char	result[9];

	if(l > 8)l = 8;
	if(l < 1)l = 1;
	result[l] = 0;
	result[--l] = hexnum[c % 0x10];
	while( l-- )
		result[l] = hexnum[(c >>= 4) % 0x10];
	fputs(result, fp);
}

static	unsigned long	makelong(unsigned char *buffer)
{
	return	( ((long)buffer[0] << 24 ) | 
	      	  ((long)buffer[1] << 16 ) | 
	      	  ((long)buffer[2] <<  8 ) | 
	      	  ((long)buffer[3] ) );
}

static	unsigned short	makeshort(unsigned char *buffer)
{
	return	( ((long)buffer[0] <<  8 ) | 
	      	  ((long)buffer[1] ) );
}

static	void	divlong(unsigned char *buffer, unsigned long data)
{
	buffer[0] = (unsigned char)(( data >> 24 ) & 0x000000ff );
	buffer[1] = (unsigned char)(( data >> 16 ) & 0x000000ff );
	buffer[2] = (unsigned char)(( data >>  8 ) & 0x000000ff );
	buffer[3] = (unsigned char)(  data         & 0x000000ff );
}

static	void	divshort(unsigned char *buffer, unsigned short data)
{
	buffer[0] = (unsigned char)(( data >>  8 ) & 0x000000ff );
	buffer[1] = (unsigned char)(  data         & 0x000000ff );
}
