/* Divide AppleDouble into AppleDouble */
/*          Copyright(c)2003 by T.Toda */
#include <stdio.h>
#include <string.h>

#ifdef	MSDOS
#define	READMODE	"rb"
#define	WRITEMODE	"wb"
#else
#ifdef	WIN32
#define	READMODE	"rb"
#define	WRITEMODE	"wb"
#else
#define	READMODE	"r"
#define	WRITEMODE	"w"
#endif
#endif

/* Offset of HFS Date from AppleDouble Date       */
/* Value of Seconds of 96 years from 1904 to 2000 */
#define	DATEOFFSET	((96UL / 4) * (365UL * 4 + 1) * 24UL * 60UL * 60UL)

static	void	fputhex(unsigned long, int, FILE *);
static	unsigned long	makelong(unsigned char *);
static	unsigned short	makeshort(unsigned char *);
static	void	divlong(unsigned char *, unsigned long);
static	void	divshort(unsigned char *, unsigned short);
static	unsigned short	makecrc(unsigned char *, unsigned short);

int	main(int argc, char **argv, char **envp)
{
	FILE	*fphead, *fpdata, *fpin;
	struct{
		unsigned char	oldversion;
		unsigned char	fnamelen;
		char	filename[63];
		char	filetype[4];
		char	creator[4];
		unsigned char	finderflag1;
		unsigned char	zerofill1;
		unsigned char	vertical[2];
		unsigned char	horizontal[2];
		unsigned char	folderid[2];
		unsigned char	protected;
		unsigned char	zerofill2;
		unsigned char	datafolk[4];
		unsigned char	resourcefolk[4];
		unsigned char	created[4];
		unsigned char	modified[4];
		unsigned char	infocomment[2];	/* for MacBinary II */
		unsigned char	finderflag2;	/* for MacBinary II */
		unsigned char	signature[4];	/* for MacBinary III */
		unsigned char	fnamescript;	/* for MacBinary III */
		unsigned char	extendedflag;	/* for MacBinary III */
		unsigned char	reserved[8];
		unsigned char	unpackedsize[4];	/* for MacBinary II */
		unsigned char	secondheader[2];	/* for MacBinary II */
		unsigned char	version;	/* for MacBinary II */
		unsigned char	minversion;	/* for MacBinary II */
		unsigned char	crc[2];	/* for MacBinary II */
		unsigned char	osid[2];	/* for MacBinary I */
	}	macbinheader;
	struct{
		unsigned char	magic[4];
		unsigned char	version[4];
		char	filler[16];
		unsigned char	entries[2];
	}	appleheader;
	struct{
		unsigned char	id[4];
		unsigned char	offset[4];
		unsigned char	length[4];
	}	appleentry;
	unsigned long	lastoffset, l;
	unsigned short	crc_rcrd, crc_calc;
	int	c;

	if( argc > 4 || argc < 3 ){
		fputs("Usage : ", stderr);
		fputs(argv[0], stderr);
		fputs(" infile headerfile [datafile]\n", stderr);
		return	4;
	}
	if( (fpin = fopen(argv[1], READMODE)) == NULL ){
		fputs("Can't Open '", stderr);
		fputs(argv[1], stderr);
		fputs("'.\n", stderr);
		return	8;
	}
	if( (fphead = fopen(argv[2], READMODE)) != NULL ){
		fputs("Destinate File '", stderr);
		fputs(argv[2], stderr);
		fputs("' Exists. Overwrite? (Y/N)", stderr);
		c = getchar();
		if( (c != 'Y') && (c != 'y') ){
			return	8;
		}
		while( getchar() != '\n' );
	}
	fclose(fphead);
	if( (fphead = fopen(argv[2], WRITEMODE)) == NULL ){
		fputs("Can't Open '", stderr);
		fputs(argv[2], stderr);
		fputs("'.\n", stderr);
		return	8;
	}
	if( argc < 4 ){
		fpdata = stdout;
	}
	else{
		if( (fpdata = fopen(argv[3], READMODE)) != NULL ){
			fputs("Destinate File '", stderr);
			fputs(argv[3], stderr);
			fputs("' Exists. Overwrite? (Y/N)", stderr);
			c = getchar();
			if( (c != 'Y') && (c != 'y') ){
				return	8;
			}
			while( getchar() != '\n' );
		}
		fclose(fpdata);
		if( (fpdata = fopen(argv[3], WRITEMODE)) == NULL ){
			fputs("Can't Open '",stderr);
			fputs(argv[3],stderr);
			fputs("'.\n",stderr);
			return	8;
		}
	}

	fread(&macbinheader, sizeof(macbinheader), 1, fpin);
	crc_rcrd = makeshort(macbinheader.crc);
	divshort(macbinheader.crc, 0);
	crc_calc = makecrc((unsigned char *)&macbinheader, 0x7e);
	if( crc_calc != crc_rcrd ){
		fputs("CRC for Header Mismatch: Calculated = ", stderr);
		fputhex((unsigned long)crc_calc, 4, stderr);
		fputs(", Recorded = ", stderr);
		fputhex((unsigned long)crc_rcrd, 4, stderr);
		putc('\n', stderr);
	}

	memset(&appleheader, 0, sizeof(appleheader));
	lastoffset = sizeof(appleheader);
	lastoffset += sizeof(appleentry) * 5;
	divshort(appleheader.entries, 5);
		/* 5 = Name, Finder, Date, Comment, RF (no DF) */
	divlong(appleheader.magic, 0x00051607);
	divlong(appleheader.version, 0x00020000);
	fwrite(&appleheader, sizeof(appleheader), 1, fphead);

	/* 3: Header for Real Name */
	divlong(appleentry.id, 3L);
	divlong(appleentry.length, (unsigned long)macbinheader.fnamelen);
	divlong(appleentry.offset, lastoffset);
	lastoffset += macbinheader.fnamelen;
	fwrite(&appleentry, sizeof(appleentry), 1, fphead);

	/* 9: Header for Finder Info */
	divlong(appleentry.id, 9L);
	divlong(appleentry.length, 32L);
	divlong(appleentry.offset, lastoffset);
	lastoffset += 32L;
	fwrite(&appleentry, sizeof(appleentry), 1, fphead);

	/* 8: Header for File Dates */
	divlong(appleentry.id, 8L);
	divlong(appleentry.length, 16L);
	divlong(appleentry.offset, lastoffset);
	lastoffset += 16L;
	fwrite(&appleentry, sizeof(appleentry), 1, fphead);

	/* 2: Header for Resource Fork */
	divlong(appleentry.id, 2L);
	memcpy(appleentry.length, macbinheader.resourcefolk, 4);
	divlong(appleentry.offset, lastoffset);
	lastoffset += makelong(macbinheader.resourcefolk);
	fwrite(&appleentry, sizeof(appleentry), 1, fphead);

	/* 4: Header for Comment */
	divlong(appleentry.id, 4L);
	divlong(appleentry.length, (unsigned long)makeshort(macbinheader.infocomment));
	divlong(appleentry.offset, lastoffset);
	lastoffset += makeshort(macbinheader.infocomment);
	fwrite(&appleentry, sizeof(appleentry), 1, fphead);

	/* 3: Body of Real Name */
	fwrite(macbinheader.filename, macbinheader.fnamelen, 1, fphead);

	/* 9: Body of Finder Info */
	/* Consists of 16-Byte 'Finder Info' and      */
	/* 16-Byte 'Exteneded Finder Info' (ignored). */
	/* 'Finder Info' consists of                  */
	/* 'filetype', 'creator', 'finderflag1',      */
	/* 'finderflag1', 'zerofill1' ('finderflag2') */
	/* 'vertical', 'horizontal', and 'folderid'.  */
	/* 'finderflag2' is added in MacBinary II and */
	/* its counterpart 'zerofill1' should be      */
	/* filled with 0 for compatibility.           */
	macbinheader.zerofill1 = macbinheader.finderflag2;
	fwrite(macbinheader.filetype, 16, 1, fphead);

	/*            Extended Finder Info */
	memset(macbinheader.filetype, 0, 16);
	fwrite(macbinheader.filetype, 16, 1, fphead);

	/* 8: Body of File Dates */
	divlong(macbinheader.created, makelong(macbinheader.created) - DATEOFFSET);
	divlong(macbinheader.modified, makelong(macbinheader.modified) - DATEOFFSET);
	fwrite(macbinheader.created, 8, 1, fphead);

	/*            Backup and Access Dates */
	divlong(macbinheader.created, 0x80000000);
	divlong(macbinheader.modified, 0x80000000);
	fwrite(macbinheader.created, 8, 1, fphead);

	/* 1: Body of Data Fork */
	l = makelong(macbinheader.datafolk);
	while( l-- ){
		if( (c = getc(fpin)) == EOF ){
			fputs("Error: End of File in Estimated Data Fork\n", stderr);
			return	12;
		}
		putc(c, fpdata);
	}
	l = makelong(macbinheader.datafolk);
	while( l % 128 ){
		getc(fpin);
		l++;
	}

	/* 2: Body of Resource Fork */
	l = makelong(macbinheader.resourcefolk);
	while( l-- ){
		if( (c = getc(fpin)) == EOF ){
			fputs("Error: End of File in Estimated Resource Fork\n", stderr);
			return	12;
		}
		putc(c, fphead);
	}
	l = makelong(macbinheader.resourcefolk);
	while( l % 128 ){
		getc(fpin);
		l++;
	}

	/* 4: Body of Comment */
	l = makeshort(macbinheader.infocomment);
	while( l-- ){
		if( (c = getc(fpin)) == EOF ){
			fputs("Error: End of File in Estimated GetInfo Comment\n", stderr);
			return	12;
		}
		putc(c, fphead);
	}

	return	0;
}

#define	CRC16_POLYNOMIAL	0x1021;
static	unsigned short	makecrc(unsigned char *dataseq, unsigned short length)
{
	unsigned short	result = 0;
	unsigned short	i, byte, bit;

	for( i = 0 ; i < length ; i++ ){
		byte = dataseq[i];
		for( bit = 0 ; bit < 8 ; bit++ ){
			if( result & 0x8000 ){
				result <<= 1;
				result ^= CRC16_POLYNOMIAL;
			}
			else{
				result <<= 1;
			}
			if( byte & 0x80 ){
				result ^= 1;
			}
			byte <<= 1;
		}
	}
	return	result;
}

static void	fputhex(unsigned long c, int l, FILE *fp)
{
	static	char	const	hexnum[] = "0123456789ABCDEF";
	char	result[9];

	if(l > 8)l = 8;
	if(l < 1)l = 1;
	result[l] = 0;
	result[--l] = hexnum[c % 0x10];
	while( l-- )
		result[l] = hexnum[(c >>= 4) % 0x10];
	fputs(result, fp);
}

static	unsigned long	makelong(unsigned char *buffer)
{
	return	( ((long)buffer[0] << 24 ) | 
	      	  ((long)buffer[1] << 16 ) | 
	      	  ((long)buffer[2] <<  8 ) | 
	      	  ((long)buffer[3] ) );
}

static	unsigned short	makeshort(unsigned char *buffer)
{
	return	( ((long)buffer[0] <<  8 ) | 
	      	  ((long)buffer[1] ) );
}

static	void	divlong(unsigned char *buffer, unsigned long data)
{
	buffer[0] = (unsigned char)(( data >> 24 ) & 0x000000ff );
	buffer[1] = (unsigned char)(( data >> 16 ) & 0x000000ff );
	buffer[2] = (unsigned char)(( data >>  8 ) & 0x000000ff );
	buffer[3] = (unsigned char)(  data         & 0x000000ff );
}

static	void	divshort(unsigned char *buffer, unsigned short data)
{
	buffer[0] = (unsigned char)(( data >>  8 ) & 0x000000ff );
	buffer[1] = (unsigned char)(  data         & 0x000000ff );
}
